package CH.ifa.draw;

import CH.ifa.draw.application.DrawApplication;
import CH.ifa.draw.application.FileMenuCreator;
import CH.ifa.draw.application.MenuManager;
import CH.ifa.draw.application.NoGuiAvailableException;

import CH.ifa.draw.framework.Drawing;
import CH.ifa.draw.framework.DrawingEditor;

import CH.ifa.draw.io.ExportHolder;
import CH.ifa.draw.io.ExportHolderImpl;
import CH.ifa.draw.io.ImportHolder;
import CH.ifa.draw.io.ImportHolderImpl;

import CH.ifa.draw.standard.AlignmentMenuCreator;
import CH.ifa.draw.standard.AttributesMenuCreator;
import CH.ifa.draw.standard.EditMenuCreator;
import CH.ifa.draw.standard.NullDrawingEditor;

import CH.ifa.draw.util.GUIProperties;

import de.renew.plugin.PluginAdapter;
import de.renew.plugin.PluginException;
import de.renew.plugin.PluginManager;
import de.renew.plugin.PluginProperties;

import java.awt.Toolkit;

import java.net.URL;

import java.util.Collection;
import java.util.Iterator;

import javax.swing.JMenuItem;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;


/**
 * This class represents the Plugin for the JHotDraw Component.
 * It creates the standard menu items provided by that framework.
 * Note that the drawing window itself is not created here yet but by the
 * GuiPlugin. This should be changed, but requires quite massive
 * restructuring.
 *
 * @author Joern Schumacher
 */
public class DrawPlugin extends PluginAdapter {
    public static org.apache.log4j.Logger logger = org.apache.log4j.Logger
                                                   .getLogger(DrawPlugin.class);
    private static DrawApplication _gui;

    /**
     * All import/export work is delegated to these objects.
     **/
    private ImportHolderImpl importHolder;

    /**
     * All import/export work is delegated to these objects.
     **/
    private ExportHolderImpl exportHolder;

    // TODO: this is a dirty hack and should be redone
    // so that the instance is controlled by this plugin only 
    public static void setGui(DrawApplication gui) {
        logger.debug("IfaPlugin setting gui to " + gui);
        _gui = gui;
    }

    public static DrawApplication getGui() {
        return _gui;
    }

    /**
     * The name of the plugins menu.
     * Extensions of small plugins with functionality not
     * matching any of the other menus can be registered here.
     **/
    public static final String PLUGINS_MENU = "Plugins";

    /**
     * The name of the file menu.
     * Commands to load, save and print drawings be registered here.
     **/
    public static final String FILE_MENU = "File";

    /**
     * The name of the edit menu.
     * Basic editing commands (selection, cut and paste, etc.) be registered here.
     **/
    public static final String EDIT_MENU = "Edit";

    /**
     * The name of the attributes menu.
     * Commands that modify figure attributes be registered here.
     **/
    public static final String ATTRIBUTES_MENU = "Attributes";

    /**
     * The name of the layout menu.
     * Commands that align or order figures be registered here.
     **/
    public static final String LAYOUT_MENU = "Layout";

    /**
     * Reserved name for the windows menu in the menu bar. This menu is
     * always placed at the end of the menu bar, but before the help menu.
     * <p>
     * This menu is implemented by {@link CH.ifa.draw.application.WindowsMenu}
     * and displays a list of drawing windows (grouped by categories)
     * instead of editing commands. This menu can also be extended
     * by plugins, as long as they do not meddle with the drawing list.
     * </p>
     **/
    public static final String WINDOWS_MENU = "Windows";

    /**
     * The name of the tools menu.
     * Commands that align or order figures be registered here.
     **/
    public static final String TOOLS_MENU = "Tools";

    /**
     * The category to use in the windows menu for tools and dialogs.
     **/
    public static final String WINDOWS_CATEGORY_TOOLS = "Tools";

    /**
     * Reserved name for the help menu in the menu bar. This
     * menu is always placed at the end of the menu bar.
     **/
    public static final String HELP_MENU = "Help";

    /**
     * This prefix is used to identify menu entries belonging to
     * this plugin. It must not be used by other plugins.
     **/
    public static final String MENU_PREFIX = "CH.ifa.draw";

    /**
     * {@inheritDoc}
     * @throws PluginException
     *   if the instantiation fails. This especially covers the case that
     *   AWT/Swing is not available.
     **/
    public DrawPlugin(URL location) throws PluginException {
        super(location);
        checkSwing();
    }

    /**
     * {@inheritDoc}
     * @throws PluginException
     *   if the instantiation fails. This especially covers the case that
     *   AWT/Swing is not available.
     **/
    public DrawPlugin(PluginProperties props) throws PluginException {
        super(props);
        checkSwing();
    }

    /**
     * Checks whether a graphical user interface can be set up.
     * This is done by calling some static Swing method, thereby
     * triggering the initialisation of the look-and-feel-specific
     * peers. If the initialisation throws any kind of exception
     * or error, this method throws a <code>PluginException</code>.
     *
     * @throws PluginException
     *   if the initialisation of the graphical user interface
     *   failed due to any kind of problem or error.
     **/
    private static void checkSwing() throws PluginException {
        try {
            Toolkit.getDefaultToolkit();
            logger.debug("DrawPlugin: Swing initialization successful.");
        } catch (Throwable e) {
            logger.debug("DrawPlugin: Swing initialization failed: " + e);
            throw new PluginException("Could not initialize Swing, disabling DrawPlugin.",
                                      e);
        }
    }

    public void init() {
        GUIProperties.setProperties(_properties);
        if (GUIProperties.fixMenus()) {
            UIManager.getDefaults().put("MenuItemUI",
                                        "CH.ifa.draw.WorkaroundMenuItemUI");
        }

        // generate Import / Export Holder
        importHolder = new ImportHolderImpl();
        exportHolder = new ExportHolderImpl();

        registerMenuItems(FILE_MENU,
                          (new FileMenuCreator()).createMenus(importHolder,
                                                              exportHolder));
        registerMenuItems(EDIT_MENU, (new EditMenuCreator()).createMenus());
        registerMenuItems(LAYOUT_MENU,
                          (new AlignmentMenuCreator()).createMenus());
        registerMenuItems(ATTRIBUTES_MENU,
                          (new AttributesMenuCreator()).createMenus());
        // the WINDOWS_MENU is automatically included by the menu manager.
    }

    /*
     * Convenience method that registers the JMenuItems contained
     * in the given collection under the given parent menu.
     * Its primary function is to properly calculate the id value of the items.
     */
    private void registerMenuItems(String menu, Collection items) {
        Iterator it = items.iterator();
        MenuManager menuManager = MenuManager.getInstance();
        while (it.hasNext()) {
            JMenuItem item = (JMenuItem) it.next();
            if (item.getClientProperty(MenuManager.ID_PROPERTY) == null) {
                item.putClientProperty(MenuManager.ID_PROPERTY,
                                       MENU_PREFIX + "." + menu + "."
                                       + item.getText());
            }
            menuManager.registerMenu(menu, item);
        }
    }

    public IOHelper getIOHelper() {
        return IOHelper.getInstance();
    }

    public ImportHolder getImportHolder() {
        return importHolder;
    }

    public ExportHolder getExportHolder() {
        return exportHolder;
    }

    public void newDrawingViewContainer(Drawing drawing) {
        getGui().newDrawingViewContainer(drawing);
    }

    public static DrawPlugin getCurrent() {
        DrawPlugin result = null;

        Iterator it = PluginManager.getInstance()
                                   .getPluginsProviding("ch.ifa.draw").iterator();
        while (it.hasNext()) {
            Object o = it.next();
            if (o instanceof DrawPlugin) {
                return (DrawPlugin) o;
            }
        }
        return null;
    }

    public MenuManager getMenuManager() {
        return MenuManager.getInstance();
    }

    public DrawingEditor getDrawingEditor() {
        DrawingEditor result = null;
        DrawApplication gui = getGui();
        if (gui == null) {
            result = NullDrawingEditor.INSTANCE;
        } else {
            result = gui;
        }
        return result;
    }

    /**
     * Displays a message in the gui status line.
     *
     * @param status  the message to show.
     * @throws NoGuiAvailableException  if the Gui window is not open.
     * @throws NullPointerException  if <code>net</code> is null.
     **/
    public void showStatus(final String status) throws NoGuiAvailableException {
        if (isGuiPresent()) {
            SwingUtilities.invokeLater(new Runnable() {
                    public void run() {
                        DrawApplication gui = getGui();
                        if (gui != null) {
                            gui.showStatus(status);
                        }
                    }
                });
        } else {
            throw new NoGuiAvailableException("DrawPlugin: cannot show status: no gui window.");
        }
    }

    /**
     * Checks whether the gui window is open.
     * Use this test to avoid NoGuiAvailableExceptions in methods
     * that depend on the Frame to be open.
     *
     * @return True if the window is active, false otherwise.
     */
    public boolean isGuiPresent() {
        return _gui != null;
    }
}