package CH.ifa.draw.standard;

import CH.ifa.draw.DrawPlugin;

import CH.ifa.draw.application.MenuManager;

import CH.ifa.draw.framework.DrawingView;
import CH.ifa.draw.framework.Figure;
import CH.ifa.draw.framework.FigureEnumeration;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.GraphicsEnvironment;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JColorChooser;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.ListSelectionModel;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;


/**
 * Command to change a font.
 * <p>
 * The command's effects are undoable step by step.
 * Each use of the apply button can be undone separately.
 * So this command doesn't need to inherit UndoableCommand.
 * </p>
 * @author Sven Offermann
 */
public class ChooseFontCommand extends QueryAttributeCommand {
    private JList fontList;
    private JList sizeList;
    private JCheckBox cbBold;
    private JCheckBox cbItalic;
    private JTextArea txtSample;
    private String[] sizes = new String[] { "2", "4", "6", "8", "10", "12", "14", "16", "18", "20", "22", "24", "30", "36", "48", "72" };
    private JFrame lastParent = null;
    private JDialog dialog = null;
    private JColorChooser chooser;

    /**
     * Constructs a dialog with a FontChooser to
     * change the font of inscriptions etc.
     *
     * @param displayName the dialog name
     * @param name the command name
     * @param attributeName the name of the attribute to be changed
     * @param type the attribute type (class)
     */
    public ChooseFontCommand(String displayName, String name,
                             String attributeName, Class type) {
        super(displayName, name, attributeName, type);
    }

    private JDialog getDialog() {
        // We can reuse an existing dialog, but we must create a new dialog
        // when the gui frame has been closed and recreated inbetween.
        // This is possible because the menu command objects are kept across
        // gui instances by the MenuManager
        // TODO: Redesign of the MenuManager or apply the same mechanism to
        //       all other editor dialogs (see bug #99).
        JFrame parent = DrawPlugin.getGui().getFrame();
        if (dialog != null) {
            if (lastParent == parent) {
                // reuse the existing dialog
                return dialog;
            } else {
                // or release the old one
                dialog.dispose();
            }
        }
        dialog = new JDialog(parent, "Change " + displayName + ":");

        lastParent = parent;


        // create all components
        fontList = new JList(GraphicsEnvironment.getLocalGraphicsEnvironment()
                                                .getAvailableFontFamilyNames()) {
                public Dimension getPreferredScrollableViewportSize() {
                    return new Dimension(150, 144);
                }
            };
        fontList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);

        sizeList = new JList(sizes) {
                public Dimension getPreferredScrollableViewportSize() {
                    return new Dimension(25, 144);
                }
            };
        sizeList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);

        cbBold = new JCheckBox("Bold");

        cbItalic = new JCheckBox("Italic");

        txtSample = new JTextArea() {
                public Dimension getPreferredScrollableViewportSize() {
                    return new Dimension(385, 80);
                }
            };
        txtSample.setText("This is a sample text.");


        JButton apply = new JButton(" Apply ");
        apply.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    apply(getType());
                }
            });

        JButton update = new JButton(" Update ");
        update.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    updateFromFigure();
                }
            });

        JButton ok = new JButton(" OK ");
        ok.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    apply(getType());
                    QueryAttributeCommand.closeDialog(dialog);
                }
            });

        JButton cancel = new JButton(" Cancel ");
        cancel.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    QueryAttributeCommand.closeDialog(dialog);
                }
            });

        JPanel panel = new JPanel();
        panel.setLayout(new FlowLayout());
        panel.add(apply);
        panel.add(update);
        panel.add(ok);
        panel.add(cancel);


        // set the default font
        setFont(null);

        // add the listeners
        ListSelectionListener listListener = new ListSelectionListener() {
            public void valueChanged(ListSelectionEvent e) {
                txtSample.setFont(getCurrentFont());
            }
        };

        fontList.addListSelectionListener(listListener);
        sizeList.addListSelectionListener(listListener);


        ActionListener cbListener = new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                txtSample.setFont(getCurrentFont());
            }
        };

        cbBold.addActionListener(cbListener);
        cbItalic.addActionListener(cbListener);

        // build the container
        dialog.getContentPane().setLayout(new java.awt.BorderLayout());

        JPanel fontChoosePanel = new JPanel();
        fontChoosePanel.setLayout(new BorderLayout());

        JPanel leftPanel = new JPanel();
        leftPanel.setLayout(new java.awt.BorderLayout());

        leftPanel.add(new JScrollPane(fontList), java.awt.BorderLayout.CENTER);
        leftPanel.add(new JScrollPane(sizeList), java.awt.BorderLayout.EAST);

        fontChoosePanel.add(leftPanel, java.awt.BorderLayout.CENTER);


        JPanel rightPanel = new JPanel();
        rightPanel.setLayout(new java.awt.FlowLayout());

        rightPanel.add(cbBold);
        rightPanel.add(cbItalic);

        fontChoosePanel.add(rightPanel, java.awt.BorderLayout.EAST);

        fontChoosePanel.add(new JScrollPane(txtSample),
                            java.awt.BorderLayout.SOUTH);

        dialog.getContentPane().add(fontChoosePanel, BorderLayout.CENTER);
        dialog.getContentPane().add(panel, BorderLayout.SOUTH);

        dialog.setSize(200, 200);
        dialog.setResizable(true);

        dialog.setDefaultCloseOperation(JDialog.DO_NOTHING_ON_CLOSE);
        dialog.addWindowListener(new WindowAdapter() {
                public void windowClosing(WindowEvent e) {
                    QueryAttributeCommand.closeDialog(dialog);
                }
            });

        dialog.pack();
        return dialog;
    }

    private void setFont(Font font) {
        if (font == null) {
            font = txtSample.getFont();
        }

        fontList.setSelectedValue(font.getName(), true);
        fontList.ensureIndexIsVisible(fontList.getSelectedIndex());
        sizeList.setSelectedValue("" + font.getSize(), true);
        sizeList.ensureIndexIsVisible(sizeList.getSelectedIndex());

        cbBold.setSelected(font.isBold());
        cbItalic.setSelected(font.isItalic());
    }

    private Font getCurrentFont() {
        String fontFamily = (String) fontList.getSelectedValue();
        int fontSize = Integer.parseInt((String) sizeList.getSelectedValue());

        int fontType = Font.PLAIN;

        if (cbBold.isSelected()) {
            fontType += Font.BOLD;
        }
        if (cbItalic.isSelected()) {
            fontType += Font.ITALIC;
        }

        return new Font(fontFamily, fontType, fontSize);
    }

    public void execute() {
        JDialog dialog = getDialog();

        updateFromFigure();

        dialog.show();
        MenuManager.getInstance().getWindowsMenu().addDialog(QueryAttributeCommand.WINDOWS_CATEGORY_ATTRIBUTES,
                                                             dialog);
    }

    private void updateFromFigure() {
        DrawingView view = getEditor().view();
        Object value = null;
        Font font = null;
        FigureEnumeration k = view.selectionElements();
        while ((k.hasMoreElements()) && (font == null)) {
            Figure f = k.nextFigure();
            Object fName = f.getAttribute("FontName");
            Object fStyle = f.getAttribute("FontStyle");
            Object fSize = f.getAttribute("FontSize");

            if ((fName != null) && (fStyle != null) && (fSize != null)) {
                font = new Font((String) fName, ((Integer) fStyle).intValue(),
                                ((Integer) fSize).intValue());
            }
        }

        setFont(font);
    }

    public boolean isExecutable() {
        if (getEditor() == NullDrawingEditor.INSTANCE) {
            return false;
        }
        return getEditor().view().selectionCount() > 0;
    }

    private boolean apply(Class type) {
        String fontFamily = (String) fontList.getSelectedValue();
        Integer fontSize = new Integer((String) sizeList.getSelectedValue());

        int fontType = Font.PLAIN;
        if (cbBold.isSelected()) {
            fontType += Font.BOLD;
        }
        if (cbItalic.isSelected()) {
            fontType += Font.ITALIC;
        }

        // set the new font attributes of the selected figures
        // TODO: Do the attribute changes in a new ChangeAttributeCommand,
        //       which changes the three font attributes in one step, so
        //       the undo function works better.
        new ChangeAttributeCommand("", "FontName", fontFamily).execute();
        new ChangeAttributeCommand("", "FontSize", fontSize).execute();
        new ChangeAttributeCommand("", "FontStyle", new Integer(fontType))
        .execute();

        return true;
    }
}