package CH.ifa.draw.util;

import java.util.Vector;


/**
 * Additional Graphics methods, e.g. for drawing dashed lines.
 **/
public interface ExtendedGraphics {

    /**
     * Draws a dashed line starting at (x1,y1) and ending at (x2,y2)
     * with the interval size <code>strokeSize</code>. That means a
     * stroke of length <code>strokeSize</code> is followed by a gap of
     * length <code>strokeSize</code> and so on.
     *
     * @param strokeSize the length of dashes and gaps (as described
     *                   above).
     * @param x1         the first point's x coordinate.
     * @param y1         the first point's y coordinate.
     * @param x2         the second point's x coordinate.
     * @param y2         the second point's y coordinate.
     *
     * @see java.awt.Graphics#drawLine
     */
    public void drawDashedLine(int x1, int y1, int x2, int y2, int strokeSize);

    /**
     * Draws a dashed line starting at (x1,y1) and ending at (x2,y2)
     * using the stroke pattern defined in <code>strokeSequence</code>.
     * <p>
     * The <code>strokeSequence</code> is a string of integers separated
     * by spaces. The values alternating determine dash and gap lengths,
     * beginning with a dash length. The string has to contain an even
     * number of such values and may not be empty.
     * As a special case it is allowed to supply one single value, this
     * value will then be used both for dashes and gaps.
     * <pre>
     * Examples:
     *    "1 7"       dotted line           . . . . . . .
     *
     *    "5 4"       dashed line           _ _ _ _ _ _ _
     *
     *    "5 4 1 4"   dash-dotted line      _ . _ . _ . _
     *
     *    "4"         dashed line           - - - - - - -
     * </pre>
     * If you want to check a stroke pattern for validity, use the
     * static method {@link ExtendedGraphicsImpl#isValidStrokePattern}.
     * </p>
     *
     * @param strokeSequence the pattern for dashes and gaps (as
     *                   described above).
     * @param x1         the first point's x coordinate.
     * @param y1         the first point's y coordinate.
     * @param x2         the second point's x coordinate.
     * @param y2         the second point's y coordinate.
     *
     * @see java.awt.Graphics#drawLine
     */
    public void drawDashedLine(int x1, int y1, int x2, int y2,
                               String strokeSequence);

    /**
     * Draws a dashed polyline with equal dash and gap
     * lengths. The nodes of the polyline are given by
     * <code>int</code> arrays in the same way as
     * {@link java.awt.Graphics#drawPolyline} needs them.
     *
     * @param strokeSize the length of dashes and gaps
     *                   (see {@link #drawDashedLine(int,int,int,int,int)}).
     * @param xPoints    an array of x coordinates.
     * @param yPoints    an array of y coordinates.
     * @param nPoints    the total number of points.
     *
     * @see #drawDashedLine(int,int,int,int,int)
     * @see java.awt.Graphics#drawPolyline
     */
    public void drawDashedPolyline(int[] xPoints, int[] yPoints, int nPoints,
                                   int strokeSize);

    /**
     * Draws a dashed polyline with variable dash and gap
     * lengths. The nodes of the polyline are given by
     * <code>int</code> arrays in the same way as
     * {@link java.awt.Graphics#drawPolyline} needs them.
     *
     * @param strokeSequence the pattern for dashes and gaps
     *                   (see {@link #drawDashedLine(int,int,int,int,String)}).
     * @param xPoints    an array of x coordinates.
     * @param yPoints    an array of y coordinates.
     * @param nPoints    the total number of points.
     *
     * @see #drawDashedLine(int,int,int,int,String)
     * @see java.awt.Graphics#drawPolyline
     */
    public void drawDashedPolyline(int[] xPoints, int[] yPoints, int nPoints,
                                   String strokeSequence);

    /**
     * Draws a dashed polyline with equal dash and gap
     * lengths.
     *
     * @param strokeSize the length of dashes and gaps
     *                   (see {@link #drawDashedLine(int,int,int,int,int)}).
     * @param points     a <code>Vector</code> of {@link java.awt.Point}
     *                   objects that describe the polyline.
     *
     * @see #drawDashedLine(int,int,int,int,int)
     * @see java.awt.Graphics#drawPolyline
     */
    public void drawDashedPolyline(Vector points, int strokeSize);

    /**
     * Draws a dashed polyline with variable dash and gap
     * lengths.
     *
     * @param strokeSequence the pattern for dashes and gaps
     *                   (see {@link #drawDashedLine(int,int,int,int,String)}).
     * @param points     a <code>Vector</code> of {@link java.awt.Point}
     *                   objects that describe the polyline.
     *
     * @see #drawDashedLine(int,int,int,int,String)
     * @see java.awt.Graphics#drawPolyline
     */
    public void drawDashedPolyline(Vector points, String strokeSequence);
}