package de.renew.engine.simulator;

import de.renew.database.TransactionSource;

import de.renew.engine.common.StepIdentifier;
import de.renew.engine.searcher.Searchable;
import de.renew.engine.searcher.Searcher;
import de.renew.engine.searchqueue.SearchQueue;


public class SequentialSimulator implements Simulator, Runnable {
    public static org.apache.log4j.Logger logger = org.apache.log4j.Logger
                                                   .getLogger(SequentialSimulator.class);
    private static long runCounter = 0;
    private long simulationRunId;
    private long cycle = 0L;
    private ExecuteFinder finder;
    private Searcher searcher = new Searcher();
    private boolean stepFired = false;
    private boolean isAlive = false;
    private boolean wantBreak = false;
    private Thread runThread = null;

    /**
     * If true, the simulator will wait for the event queue
     * before every firing.
     **/
    private boolean wantEventQueueDelay;

    public SequentialSimulator() {
        this(true);
    }

    public SequentialSimulator(boolean wantEventQueueDelay) {
        this.wantEventQueueDelay = wantEventQueueDelay;

        // create a new simulationRunId
        long simulationRunId = ((long) getClass().getName().hashCode() << 32)
                               + runCounter++;

        // If possible, prepare a new binding.
        findNewBinding();
    }

    public boolean isActive() {
        return isAlive;
    }

    public synchronized void startRun() {
        // Only create a new thread if none is currently running.
        if (runThread == null) {
            // Create a new thread that will run the simulation.
            runThread = new Thread(this);


            // Run the thread in the background.
            runThread.setPriority(Thread.MIN_PRIORITY);


            // Start the thread.
            runThread.start();

            try {
                TransactionSource.simulationStateChanged(true, true);
            } catch (Exception e) {
                logger.error(e.getMessage(), e);
            }
        }
    }

    // Gently stop the simulation.
    // This may take a while, because the current search for
    // a activated binding must be completed first.
    // After the completion of this method, no thread is running
    // and runThread==null.
    public synchronized void stopRun() {
        while (runThread != null) {
            wantBreak = true;
            try {
                wait();
            } catch (InterruptedException e) {
                // This is expected. But I cannot be sure that
                // I woke up from a notification.
            }
        }

        try {
            TransactionSource.simulationStateChanged(true, false);
        } catch (Exception e) {
            logger.error(e.getMessage(), e);
        }
    }

    // Terminate the simulation once and for all.
    // Do some final clean-up and exit the thread.
    public synchronized void terminateRun() {
        // In this case we can simply call stop run, because that
        // is sufficient to terminate all running threads.
        stopRun();

        try {
            TransactionSource.simulationStateChanged(false, false);
        } catch (Exception e) {
            logger.error(e.getMessage(), e);
        }
    }

    private void fire() {
        stepFired = false;
        while (isAlive && !stepFired) {
            // Enforce flow control.
            if (wantEventQueueDelay) {
                EventQueue.awaitEmptyQueue();
            }


            // Fire the binding.
            stepFired = finder.execute(nextStepIdentifier(), false);
            findNewBinding();
        }
    }

    // To exclude concurrent access to the searcher, this method may only
    // be called from the running thread or from a
    // synchronized method that has ensured that no thread is running.
    // This method is not synchronized, because it could
    // lock up termination requests.
    // If there are no more enabled transitions, this method simply
    // returns without effect.
    private void findNewBinding() {
        // An ExecuteFinder should not be recycled.
        finder = new ExecuteFinder();


        // Try all transitions that might be enabled.
        // If a successful binding was found (though not
        // neccessarily accepted), requeue this transition
        // at the end of the search.
        // If no bindings are possible, forget about this 
        // transition. A change to a relevant place will
        // trigger its reinsertion.
        while (!finder.isCompleted() && !SearchQueue.isTotallyEmpty()) {
            // Get a searchable.
            Searchable searchable = SearchQueue.extract();


            // If the searchable is not currently enabled, keep
            // track of the earliest time when it might possibly 
            // enabled.
            OverallEarliestTimeFinder timeFinder = new OverallEarliestTimeFinder(finder);

            // Perform the search. The searchable itself is used
            // as the triggerable, because it was taken from the search queue.
            searcher.searchAndRecover(timeFinder, searchable, searchable);

            // Reinsert the searchable back into the search queue,
            // if necessary.
            timeFinder.insertIntoSearchQueue(searchable);
        }


        // Did we find a binding?
        isAlive = finder.isCompleted();
    }

    // This method must be synchronized to avoid concurrent
    // accesses to restart another thread after the stop.
    public synchronized int step() {
        if (runThread != null) {
            // Just stop the current thread.
            stopRun();
            if (isAlive && !stepFired) {
                // The current thread did not make a step.
                fire();
            }
        } else {
            // Try to make another step.
            fire();
        }
        if (stepFired) {
            if (isAlive) {
                return statusStepComplete;
            } else {
                return statusLastComplete;
            }
        } else {
            return statusDisabled;
        }
    }

    // This method is called in threads that are generated on each
    // run request. The simulator class cannot inherit from thread,
    // because this method might be run multiple times.
    public void run() {
        // Do have a prepared binding?
        if (!isAlive) {
            // No, let's try to find one, although there is
            // little hope that we can find it.
            findNewBinding();
        }

        // Make steps until deadlock or until a break is requested.
        while (isAlive && !wantBreak) {
            // Grant other threads a chance to run.
            // Only needed on systems without preemptive multitasking.
            Thread.yield();


            // Fire and search for a new binding.
            fire();
        }


        // Make sure that runThread and wantBreak are updated
        // atomically.
        synchronized (this) {
            // Signal the end of the run.
            runThread = null;


            // Clear the break request.
            wantBreak = false;


            // Notify waiting threads of break.
            notifyAll();
        }
    }

    public synchronized void refresh() {
        if (runThread == null) {
            // No thread is currently running. Make sure to update
            // the enabled bindings and the time manually.
            findNewBinding();
        }
    }

    /**
     * @return <code>true</code>
     **/
    public boolean isSequential() {
        return true;
    }

    public StepIdentifier nextStepIdentifier() {
        return new StepIdentifier(simulationRunId, new long[] { ++this.cycle });
    }

    public StepIdentifier currentStepIdentifier() {
        return new StepIdentifier(simulationRunId, new long[] { this.cycle });
    }
}